"""Domain entities for post-processing."""

import dataclasses
from codecs import Codec
from enum import Enum

import ocf_blosc2


class CodecOptions(Codec, Enum):
    """Options for compression codecs."""
    UNSET = None
    OCF_BLOSC2 = ocf_blosc2.Blosc2(clevel=5)
    """Use the OCF Blosc2 codec.

    See Also:
        - https://pypi.org/project/ocf-blosc2/
    """

    def __bool__(self) -> bool:
        """Boolean indicating whether a codec is set."""
        return self != CodecOptions.UNSET


@dataclasses.dataclass(slots=True)
class PostProcessOptions:
    """Options for post-processing NWP data.

    The defaults for any option should be the null value,
    i.e. nothing occurs by default.
    """

    validate: bool = False
    """Whether to validate the data.

    Note that for the moment, this is a very memory-intensive operation.
    Turn on only if there exists RAM to spare!
    """

    codec: CodecOptions = CodecOptions.UNSET
    """Whether to compress the data with a non-standard codec.

    By default, Zarr writes chunks compressed using the `Blosc compressor
    <https://zarr.readthedocs.io/en/stable/tutorial.html#compressors>`_.
    """

    plot: bool = False
    """Whether to save a plot of the data."""


    def requires_rewrite(self) -> bool:
        """Boolean indicating whether the specified options necessitate a rewrite."""
        return any(
            [
                self.codec,
            ],
        )

    def requires_postprocessing(self) -> bool:
        """Boolean indicating whether the specified options necessitate post-processing."""
        return any(
            [
                self.validate,
                self.codec,
                self.plot,
            ],
        )




import unittest

import numpy as np
import xarray as xr
from hypothesis import given
from hypothesis import strategies as st
from returns.pipeline import is_successful

from .parameters import Parameter


class TestParameters(unittest.TestCase):
    """Test the business methods of the Parameters class."""

    @given(st.sampled_from(Parameter))
    def test_metadata(self, p: Parameter) -> None:
        """Test the metadata method."""
        metadata = p.metadata()
        self.assertEqual(metadata.name, p.value)

    @given(st.sampled_from([s for p in Parameter for s in p.metadata().alternative_shortnames]))
    def test_try_from_shortname(self, shortname: str) -> None:
        """Test the try_from_shortname method."""
        p = Parameter.try_from_alternate(shortname)
        self.assertTrue(is_successful(p))

        p = Parameter.try_from_alternate("invalid")
        self.assertFalse(is_successful(p))

    @given(
        st.sampled_from([s for p in Parameter for s in p.metadata().alternative_shortnames]),
        st.sampled_from(Parameter),
    )
    def test_rename_else_drop_ds_vars(self, shortname: str, parameter: Parameter) -> None:
        """Test the rename_else_drop_ds_vars method."""
        allowed_parameters: list[Parameter] = [parameter]

        ds = xr.Dataset(
            data_vars={
                shortname: (
                    ("init_time", "step", "latitude", "longitude"), np.random.rand(1, 12, 15, 15),
                ),
                "unknown-parameter": (
                    ("init_time", "step", "latitude", "longitude"), np.random.rand(1, 12, 15, 15),
                ),
            },
            coords={
                "init_time": np.array([0]),
                "step": np.array(range(12)),
                "latitude": np.array(range(15)),
                "longitude": np.array(range(15)),
            },
        )

        ds = Parameter.rename_else_drop_ds_vars(
            ds,
            allowed_parameters=allowed_parameters,
        )

        if shortname in parameter.metadata().alternative_shortnames:
            self.assertTrue(len(list(ds.data_vars)) == 1)
            self.assertEqual(next(iter(ds.data_vars)), str(parameter))
        else:
            self.assertTrue(len(list(ds.data_vars)) == 0)


if __name__ == "__main__":
    unittest.main()



"""Interfaces for actor-core communication.

The ports module defines abstract interfaces that specify the signatures
any actors (driving and driven) must obey in order to interact with the core.

*Driving* actors are found in the `services` module, and *driven* actors are found
in the `repositories` module.
"""

from .services import ConsumeUseCase
from .repositories import RawRepository, NotificationRepository

__all__ = [
    "ConsumeUseCase",
    "RawRepository",
    "NotificationRepository",
]



"""Model Repositories

TODO: Add description
"""

from .ceda import CEDARawRepository
from .ecmwf_realtime import ECMWFRealTimeS3RawRepository
from .ecmwf_mars import ECMWFMARSRawRepository
from .noaa_s3 import NOAAS3RawRepository
from .mo_datahub import MetOfficeDatahubRawRepository

__all__ = [
    "CEDARawRepository",
    "ECMWFRealTimeS3RawRepository",
    "NOAAS3RawRepository",
    "MetOfficeDatahubRawRepository",
    "ECMWFMARSRawRepository",
]




"""Class for tracking memory usage in a separate thread.

Adapted from the Joblib documentation:
https://joblib.readthedocs.io/en/stable/auto_examples/parallel_generator.html#memorymonitor-helper
"""

import time
from threading import Thread
from types import TracebackType

import psutil


class PerformanceMonitor(Thread):
    """Monitor the memory usage in MB in a separate thread.

    Note that this class is good enough to highlight the memory profile of
    Parallel in this example, but is not a general purpose profiler fit for
    all cases.
    """

    thread: Thread
    memory_buffer: list[int]
    cpu_buffer: list[float]
    start_time: float
    end_time: float | None
    stop: bool = True

    def __enter__(self) -> None:
        """Start the monitor."""
        super().__init__()
        self.stop = False
        self.memory_buffer: list[int] = []
        self.cpu_buffer: list[float] = []
        self.start_time = time.time()
        self.start()

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None,
        ) -> None:
        """Stop the performance monitor, saving the results."""
        self.stop = True
        self.end_time = time.time()
        super().join(timeout=30)

    def get_usage(self) -> tuple[int, float]:
        """Get usage of a process and its children."""
        p = psutil.Process()
        # CPU usage of process and its children
        cpu: float = p.cpu_percent()
        # Memory usage does not reflect child processes
        # * Manually add the memory usage of child processes
        memory: int = p.memory_info().rss
        for c in p.children():
            memory += c.memory_info().rss
        return memory, cpu

    def get_runtime(self) -> int:
        """Get the runtime of the thread in seconds."""
        if self.end_time is None:
            return int(time.time() - self.start_time)
        return int(self.end_time - self.start_time)

    def run(self) -> None:
        """Run the thread."""
        memory_start, cpu_start = self.get_usage()
        while not self.stop:
            new_memory, new_cpu = self.get_usage()
            # Memory is just a total, so get the delta
            self.memory_buffer.append(new_memory - memory_start)
            # CPU is calculated by psutil against the base CPU,
            # so no need to get a delta
            self.cpu_buffer.append(new_cpu)
            time.sleep(0.2)

    def max_memory_mb(self) -> float:
        """Get the maximum memory usage during the thread's runtime."""
        return max(self.memory_buffer) / 1e6

    def max_cpu_percent(self) -> float:
        """Get the maximum CPU usage during the thread's runtime."""
        return max(self.cpu_buffer)




"""Domain entities describing dimensional coordinates.

Multidimensional data
---------------------

Tensor datasets are the primary data structure used in the consumer, which are
characterised by their multidimensional nature. To map data points in a tensor
back to selectable, indexable points along the dimensions of the tensor, a
mapping is required between the integer ticks along the dimension axes and the
values those ticks represent.

For instance, consider a 2D tensor containing x, y data of the lap number vs lap
time of a runner running around a racetrack. The point (2, 4) in the tensor
would represent the runner's time at lap 2. In this instance the indexes are
2 and 4, but to get back to the values they represent, a mapping of the
dimension indices to coordinate values must be consulted, for instance:

x index: [0, 1, 2, 3, 4]
x value: [lap 1, lap 2, lap 3, lap 4, lap 5]

y index: [0, 1, 2, 3, 4]
y value: [0 seconds, 30 seconds, 60 seconds, 90 seconds, 120 seconds]

Now by consulting the mapping we can see that the point (2, 4) in the tensor
represents that the runners time at lap three was 60 seconds.


This formalisation is useful also in the reverse case: inserting data into a
tensor according to its dimension coordinate values, and not its indexes.
This is the primary use case for these maps in this service.

It is far more likely that for incoming data the coordinate values along the
dimension axes are known, as opposed to the indexes they represent. This mapping
then enables insertion that data into the correct regions of the tensor, which is
a key part of parallel writing.
"""

import dataclasses
import datetime as dt
import json
import logging
import math
from importlib.metadata import PackageNotFoundError, version
from typing import Any

import dask.array
import numpy as np
import pandas as pd
import pytz
import xarray as xr
from returns.result import Failure, ResultE, Success

from .parameters import Parameter

try:
    __version__ = version("nwp-consumer")
except PackageNotFoundError:
    __version__ = "v?"

log = logging.getLogger("nwp-consumer")


@dataclasses.dataclass(slots=True)
class NWPDimensionCoordinateMap:
    """Container for dimensions names and their coordinate index values.

    Each public field in the container is a dimension label, and the corresponding
    value is a list of the coordinate values for each index along the dimension.

    All NWP data has an associated init time, step, and variable,
    so these dimensions names are required. Spatial coordinates however
    differ between providers and their grids, so the known spatial dimensions
    are optional (but one of them should be present!).

    For instance, most models produce data on a latitude/longitude grid,
    but others use alternative projections resulting in x/y/ grids instead.
    """

    init_time: list[dt.datetime]
    """The init times of the forecast values."""
    step: list[int]
    """The forecast step times.

    This corresponds to the horizon of the values, which is the time
    difference between the forecast initialization time and the target
    time at which the forecast data is valid.
    """
    variable: list[Parameter]
    """The variables in the forecast data."""
    ensemble_stat: list[str] | None = None
    """The relevant ensemble statistics of the forecast data."""
    ensemble_member: list[int] | None = None
    """The ensemble member numbers making up the ensemble forecast."""
    latitude: list[float] | None = None
    """The latitude coordinates of the forecast grid in degrees.

    Will be truncated to 4 decimal places, and ordered as 90 -> -90.
    """
    longitude: list[float] | None = None
    """The longitude coordinates of the forecast grid in degrees. """

    y_osgb: list[int] | None = None
    """Y coordinates of an OSGB grid."""
    x_osgb: list[int] | None = None
    """X coordinates of an OSGB grid."""

    y_laea: list[int] | None = None
    """Y coordinates of a Lambert Azimuthal Equal Area grid."""
    x_laea: list[int] | None = None
    """X coordinates of a Lambert Azimuthal Equal Area grid."""

    def __post_init__(self) -> None:
        """Rigidly set input value ordering and precision."""
        self.variable = sorted(self.variable)
        # Make latitude descending, longitude ascending, and both rounded to 4 d.p.
        # NOTE: For latitude and longitude values, we round to 4 decimal places
        # to avoid floating point precision issues when comparing values.
        # It is important to note that this places a limit on the precision
        # of the latitude and longitude values that can be stored in the map.
        # 4 decimal places corresponds to a precision of ~11m at the equator.
        if self.latitude is not None:
            self.latitude = sorted([float(f"{lat:.4f}") for lat in self.latitude], reverse=True)
        if self.longitude is not None:
            self.longitude = sorted([float(f"{lon:.4f}") for lon in self.longitude])

    @property
    def dims(self) -> list[str]:
        """Get instantiated dimensions.

        Ignores any dimensions that do not have a corresponding coordinate
        index value list.
        """
        return [f.name for f in dataclasses.fields(self) if getattr(self, f.name) is not None]

    @property
    def shapemap(self) -> dict[str, int]:
        """Mapping of dimension names to lengths; the 'shape' of the coordinates.

        This is the length of each dimension in the map,
        which can be thought of as the number of ticks along each dimension
        axis.
        """
        return {dim: len(getattr(self, dim)) for dim in self.dims}

    @property
    def coord_system(self) -> dict[str, Any]:
        """The coordinate system of the map.

        Returns:
            A string representing the coordinate system of the map.
        """
        if self.latitude is not None and self.longitude is not None:
            return {"geodesic_ellipsiodal": {"crs": "EPSG:4326"}}
        if self.y_osgb is not None and self.x_osgb is not None:
            return {
                "transverse_mercator": {
                    "latitude_of_projection_origin": 49.0,
                    "longitude_of_central_meridian": -2.0,
                    "false_easting": 400000.0,
                    "false_northing": -100000.0,
                    "scale_factor_at_central_meridian": 0.0,
                    "ellipsoid": {
                        "semi_major_axis": 6377563.4,
                        "semi_minor_axis": 6356256.91,
                    },
                },
            }
        if self.y_laea is not None and self.x_laea is not None:
            return {
                "lambert_azimuthal_equal_area": {
                    "latitude_of_projection_origin": 54.9,
                    "longitude_of_projection_origin": -2.5,
                    "false_easting": 0.0,
                    "false_northing": 0.0,
                    "ellipsoid": {
                        "semi_major_axis": 6378137.0,
                        "semi_minor_axis": 6356752.314140356,
                    },
                },
            }
        return {}

    @classmethod
    def from_pandas(
            cls,
            pd_indexes: dict[str, pd.Index], # type: ignore
    ) -> ResultE["NWPDimensionCoordinateMap"]:
        """Create a new NWPDimensionCoordinateMap from a dictionary of pandas Index objects.

        This is useful for interoperability with xarray, which prefers to define
        DataArray coordinates using a dict pandas Index objects.

        To extract the coordinate values from an xarray DataArray,
        there is the "indexes" property on an xarray Coordinates object:

        Example:
        >>> > idxs = xr_data.coords.indexes
        >>> > NWPDimensionCoordinateMap.from_pandas(idxs)
        >>> {
        >>>     "init_time": [dt.datetime(2021, 1, 1, 0, 0)],
        >>>     "step": [1, 2],
        >>>     "variable": [Parameter.TEMPERATURE_SL],
        >>>     "latitude": [90, 80, 70],
        >>>     "longitude": [45, 50, 55],
        >>> }

        See Also:
            `NWPDimensionCoordinateMap.to_pandas` for the reverse operation.
        """
        if not all(key in pd_indexes for key in ["init_time", "step", "variable"]):
            return Failure(KeyError(
                f"Cannot create {cls.__class__.__name__} instance from pandas indexes "
                "as required keys 'init_time', 'step', and 'variable' are not all present. "
                f"Got: '{list(pd_indexes.keys())}'",
            ))
        if not all(len(pd_indexes[key].to_list()) > 0 for key in ["init_time", "step", "variable"]):
            return Failure(ValueError(
                f"Cannot create {cls.__class__.__name__} instance from pandas indexes "
                "as the 'init_time', 'step', and 'variable' dimensions must have "
                "at least one coordinate value.",
            ))
        input_parameter_set: set[str] = set(pd_indexes["variable"].to_list())
        known_parameter_set: set[str] = {str(p) for p in Parameter}
        if not input_parameter_set.issubset(known_parameter_set):
            return Failure(ValueError(
                f"Cannot create {cls.__class__.__name__} instance from pandas indexes "
                "as the 'variable' dimension contains unknown parameters: ",
                f"'{list(input_parameter_set.difference(known_parameter_set))}'. "
                "Ensure the parameter names match the names of the standard parameter set "
                "defined by the `entities.Parameter` Enum.",
            ))
        if not all(key in [f.name for f in dataclasses.fields(cls)] for key in pd_indexes):
            unknown_keys: list[str] = list(
                set(pd_indexes.keys()).difference([f.name for f in dataclasses.fields(cls)]),
            )
            return Failure(KeyError(
                f"Cannot create {cls.__class__.__name__} instance from pandas indexes "
                f"as unknown index/dimension keys were encountered: {unknown_keys}.",
            ))
        if "latitude" in pd_indexes \
            and pd_indexes["latitude"].values[0] < pd_indexes["latitude"].values[-1]:
            return Failure(ValueError(
                "Cannot create NWPDimensionCoordinateMap instance from pandas indexes "
                "as the latitude values are not in descending order. "
                "Latitude coordinates should run from 90 -> -90. "
                "Modify the coordinate in the source data to be in descending order.",
            ))
        if "longitude" in pd_indexes \
            and pd_indexes["longitude"].values[0] > pd_indexes["longitude"].values[-1]:
            return Failure(ValueError(
                "Cannot create NWPDimensionCoordinateMap instance from pandas indexes "
                "as the longitude values are not in ascending order. "
                "Longitude coordinates should run from -180 -> 180. "
                "Modify the coordinate in the source data to be in ascending order.",
            ))
        if "y_osgb" in pd_indexes \
            and pd_indexes["y_osgb"].values[0] < pd_indexes["y_osgb"].values[-1]:
            return Failure(ValueError(
                "Cannot create NWPDimensionCoordinateMap instance from pandas indexes "
                "as the y_osgb values are not in descending order. "
                "Modify the coordinate in the source data to be in descending order.",
            ))
        if "x_osgb" in pd_indexes \
            and pd_indexes["x_osgb"].values[0] > pd_indexes["x_osgb"].values[-1]:
            return Failure(ValueError(
                "Cannot create NWPDimensionCoordinateMap instance from pandas indexes "
                "as the x_osgb values are not in ascending order. "
                "Modify the coordinate in the source data to be in ascending order.",
            ))
        if "y_laea" in pd_indexes \
            and pd_indexes["y_laea"].values[0] < pd_indexes["y_laea"].values[-1]:
            return Failure(ValueError(
                "Cannot create NWPDimensionCoordinateMap instance from pandas indexes "
                "as the y_laea values are not in descending order. "
                "Modify the coordinate in the source data to be in descending order.",
            ))
        if "x_laea" in pd_indexes \
            and pd_indexes["x_laea"].values[0] > pd_indexes["x_laea"].values[-1]:
            return Failure(ValueError(
                "Cannot create NWPDimensionCoordinateMap instance from pandas indexes "
                "as the x_laea values are not in ascending order. "
                "Modify the coordinate in the source data to be in ascending order.",
            ))

        # Convert the pandas Index objects to lists of the appropriate types
        return Success(
            cls(
                # NOTE: The timezone information is stripped from the datetime objects
                # as numpy cannot handle timezone-aware datetime objects. As such, it
                # must be added back in when converting to a datetime object.
                init_time=[
                    ts.to_pydatetime().replace(tzinfo=dt.UTC)
                    for ts in pd_indexes["init_time"].to_list()
                ],
                step=[np.timedelta64(ts, "h").astype(int) for ts in pd_indexes["step"].to_list()],
                # NOTE: This list comprehension can be done safely, as above we have
                # already performed a check on the pandas variable names being a subset
                # of the `Parameter` enum value names.
                variable=[
                    Parameter(pdp)
                    for pdp in pd_indexes["variable"].to_list()
                ],
                ensemble_stat=pd_indexes["ensemble_stat"].to_list() \
                    if "ensemble_stat" in pd_indexes else None,
                ensemble_member=pd_indexes["ensemble_member"].to_list() \
                    if "ensemble_member" in pd_indexes else None,
                latitude=pd_indexes["latitude"].to_list() \
                    if "latitude" in pd_indexes else None,
                longitude=pd_indexes["longitude"].to_list() \
                    if "longitude" in pd_indexes else None,
                y_osgb=pd_indexes["y_osgb"].to_list() \
                    if "y_osgb" in pd_indexes else None,
                x_osgb=pd_indexes["x_osgb"].to_list() \
                    if "x_osgb" in pd_indexes else None,
                y_laea=pd_indexes["y_laea"].to_list() \
                    if "y_laea" in pd_indexes else None,
                x_laea=pd_indexes["x_laea"].to_list() \
                    if "x_laea" in pd_indexes else None,
            ),
        )

    @classmethod
    def from_xarray(cls, xarray_obj: xr.DataArray | xr.Dataset) \
            -> ResultE["NWPDimensionCoordinateMap"]:
        """Create a new NWPDimensionCoordinateMap from an XArray DataArray or Dataset object."""
        return cls.from_pandas(xarray_obj.coords.indexes)   # type: ignore

    def to_pandas(self) -> dict[str, pd.Index]:  # type: ignore
        """Convert the coordinate map to a dictionary of pandas Index objects.

        This is useful for interoperability with xarray, which prefers to define
        DataArray coordinates using a dict pandas Index objects.

        For the most part, the conversion consists of a straightforward cast
        to a pandas Index object. However, there are some caveats involving
        the time-centric dimensions:

        - XArray will complain if any of the numpy time types have any precision
          other than nanoseconds, so care is taken to convert all time types to
          np.timedelta64['ns'] or np.datetime64['ns'] as appropriate.
        - Similarly, numpy can't handle timezone-aware datetime objects, so
          any timezone information is stripped before conversion.

        See Also:
            `NWPDimensionCoordinateMap.from_pandas` for the reverse operation.

        """
        out_dict: dict[str, pd.Index] = {  # type: ignore
            "init_time": pd.Index(
                [
                    np.datetime64(t.astimezone(pytz.utc).replace(tzinfo=None), "ns")
                    for t in self.init_time
                ],
            ),
            "step": pd.Index([np.timedelta64(np.timedelta64(h, "h"), "ns") for h in self.step]),
            "variable": pd.Index([p.value for p in self.variable]),
        } | {
            dim: pd.Index(getattr(self, dim))
            for dim in self.dims
            if dim not in ["init_time", "step", "variable"]
        }
        return out_dict

    def determine_region(
        self,
        inner: "NWPDimensionCoordinateMap",
    ) -> ResultE[dict[str, slice]]:
        """Return the index slices of inner mapping relative to the outer map.

        The caller is the "outer" dimension mapping, which the "inner" should be
        a subset of. A number of requirements must be met for this operation to be
        successful:

        - The inner must be a subset of the outer mapping along all dimensions
          (i.e. all coordinate values in the inner must be present in the outer
          for each dimension).
        - The inner's coordinate values must be contiguous along each dimension
          of the outer.
        - The inner must be of the same dimension map type as the outer map
          (i.e. must have exactly the same dimension labels).

        The returned dictionary of slices defines the region of the base map covered
        by the instances dimension mapping.

        Note that xarray does have its own implementation of this: the "region='auto'"
        argument to the "to_zarr" method performs a similar function. This is
        reimplemented in this package partly to ensure consistency of behaviour,
        partly to enable more descriptive logging in failure states, and partly to
        enable extending the functionality.

        Args:
            inner: The dimension coordinate dictionary of the smaller dataset.

        Examples:
            Getting the inner map slices relative to the outer map:

            >>> outer = NWPCoordinateMap(
            >>>     init_time=[dt.datetime(2021, 1, 1, 0, 0)],
            >>>     step=list(range(48)),
            >>>     variable=["temperature_sl", "downward_shortwave_radiation_flux_gl"],
            >>>     latitude=[68.0, 69.0, 70.0],
            >>>     longitude=[-10.0, -9.0, -8.0]
            >>> ),
            >>> inner = outer.copy()
            >>> # Modify the step of the inner to only cover half the outer
            >>> inner["step"] = list(range(24))
            >>>
            >>> outer.determine_region(inner)
            Success({
                "init_time": slice(0, 1),
                "step": slice(0, 24), # Notice the slice is inclusive of the last index
                "variable": slice(0, 2),
                "latitude": slice(0, 3),
                "longitude": slice(0, 3),
            })

        Returns:
            Dictionary mapping the slices defining the indexes of the coordinates in
            the outer dataset that correspond to the coordinates of the inner
        """
        # Ensure the inner and outer maps have the same rank and dimension labels
        if inner.dims != self.dims:
            return Failure(
                KeyError(
                    "Cannot find slices in non-matching coordinate mappings: "
                    "both objects must have identical dimensions (rank and labels). "
                    f"Got: {inner.dims} (inner) and {self.dims} (outer).",
                ),
            )

        # Ensure the inner map is entirely contained within the outer map
        slices = {}
        for inner_dim_label in inner.dims:
            inner_dim_coords = getattr(inner, inner_dim_label)
            outer_dim_coords = getattr(self, inner_dim_label)
            if len(inner_dim_coords) > len(outer_dim_coords):
                return Failure(
                    ValueError(
                        f"Coordinate values for dimension '{inner_dim_label}' in the inner map "
                        "exceed the number of coordinate values in the outer map. "
                        f"Got: {len(inner_dim_coords)} (> {len(outer_dim_coords)}) "
                        f"coordinate values.",
                    ),
                )
            if not set(inner_dim_coords).issubset(set(outer_dim_coords)):
                diff_coords = list(set(inner_dim_coords).difference(set(outer_dim_coords)))
                first_diff_index: int = inner_dim_coords.index(diff_coords[0])
                return Failure(
                    ValueError(
                        f"Coordinate values for dimension '{inner_dim_label}' not all present "
                        "within outer dimension map. The inner map must be entirely contained "
                        "within the outer map along every dimension. "
                        f"Got: {len(diff_coords)}/{len(outer_dim_coords)} differing values. "
                        f"First differing value: '{diff_coords[0]}' (inner[{first_diff_index}]) != "
                        f"'{outer_dim_coords[first_diff_index]}' (outer[{first_diff_index}]).",
                    ),
                )

            # Ensure the inner map's coordinate values are contiguous in the outer map.
            # * First, get the index of the corresponding value in the outer map for each
            #   coordinate value in the inner map:
            outer_dim_indices = sorted(
                [outer_dim_coords.index(c) for c in inner_dim_coords],
            )
            contiguous_index_run = list(range(outer_dim_indices[0], outer_dim_indices[-1] + 1))
            if outer_dim_indices != contiguous_index_run:
                idxs = np.argwhere(np.gradient(outer_dim_indices) > 1).flatten()
                return Failure(
                    ValueError(
                        f"Coordinate values for dimension '{inner_dim_label}' do not correspond "
                        f"with a contiguous index set in the outer dimension map. "
                        f"Non-contiguous values '{[outer_dim_coords[i] for i in idxs]} "
                        f"(index {[outer_dim_indices[i] for i in idxs]})' "
                        f"adjacent in dimension coordinates.",
                    ),
                )

            slices[inner_dim_label] = slice(outer_dim_indices[0], outer_dim_indices[-1] + 1)

        return Success(slices)

    def chunking(self, chunk_count_overrides: dict[str, int]) -> dict[str, int]:
        """The expected chunk sizes for each dimension.

        A dictionary mapping of dimension labels to the size of a chunk along that
        dimension. Note that the number is chunk size, not chunk number, so a chunk
        that wants to cover the entire dimension should have a size equal to the
        dimension length.

        It defaults to a single chunk per init time, step, and variable coordinate,
        and 2 chunks for each entire other dimension, unless overridden by the
        `chunk_count_overrides` argument.

        The defaults are purposefully small, to ensure that when performing parallel
        writes, chunk boundaries are not crossed.

        Args:
            chunk_count_overrides: A dictionary mapping dimension labels to the
                number of chunks to split the dimension into.
        """
        default_dict: dict[str, int] = {
            dim: 1
                if len(getattr(self, dim)) <= 8 or dim in ["init_time", "step", "variable"]
                else math.ceil(len(getattr(self, dim)))
            for dim in self.dims
        }

        out_dict = {}
        for key in default_dict:
            out_dict[key] = chunk_count_overrides.get(key, default_dict[key])

        return out_dict


    def as_zeroed_dataarray(self, name: str, chunks: dict[str, int]) -> xr.DataArray:
        """Express the coordinates as an xarray DataArray.

        The underlying dask array is a zeroed array with the shape of the dataset,
        that is chunked according to the given chunking scheme.

        Args:
            name: The name of the DataArray.
            chunks: A mapping of dimension names to the size of the chunks
                along the dimensions.

        See Also:
            - https://docs.xarray.dev/en/stable/user-guide/io.html#distributed-writes
        """
        # Create a dask array of zeros with the shape of the dataset
        # * The values of this are ignored, only the shape and chunks are used
        dummy_values = dask.array.zeros(  # type: ignore
            shape=list(self.shapemap.values()),
            chunks=tuple([chunks[k] for k in self.shapemap]),
        )
        attrs: dict[str, str] = {
            "produced_by": "".join((
                f"nwp-consumer {__version__} at ",
                f"{dt.datetime.now(tz=dt.UTC).strftime('%Y-%m-%d %H:%M')}",
            )),
            "variables": json.dumps({
                p.value: {
                    "description": p.metadata().description,
                    "units": p.metadata().units,
                } for p in self.variable
            }),
            "coord_system": json.dumps(self.coord_system),
        }
        # Create a DataArray object with the given coordinates and dummy values
        da: xr.DataArray = xr.DataArray(
            name=name,
            data=dummy_values,
            coords=self.to_pandas(),
            attrs=attrs,
        )
        return da

    def crop(
            self,
            north: float,
            west: float,
            south: float,
            east: float,
        ) -> ResultE["NWPDimensionCoordinateMap"]:
        """Return a new map cropped to the given region.

        Args:
            north: The northernmost latitude of the region in degrees.
            west: The westernmost longitude of the region in degrees.
            south: The southernmost latitude of the region in degrees.
            east: The easternmost longitude of the region in degrees.

        Returns:
            A new NWPDimensionCoordinateMap object with the latitude and longitude
            coordinates cropped to the given region.
        """
        # Ensure the region is valid
        if self.latitude is None or self.longitude is None:
            return Failure(ValueError(
                "Cannot crop coordinates to a region as latitude and/or longitude "
                "dimension coordinates are not present in the map. "
                f"Dimensions: '{self.dims}'.",
            ))

        if not (-90 <= south < north <= 90 and -180 <= west < east <= 180):
            return Failure(ValueError(
                "Cannot crop coordinates to an invalid region. "
                f"North ({north}) must be greater than South ({south}) "
                " and both must sit between 90 and -90 degrees; "
                f"East ({east}) greater than West ({west}) "
                " and both must sit between 180 and -180 degrees.",
            ))

        if (north > self.latitude[0] or south < self.latitude[-1]
            or west < self.longitude[0] or east > self.longitude[-1]):
            return Failure(ValueError(
                "Cannot crop coordinates to a region outside the bounds of the map. "
                f"Crop region '{north, west, south, east}' not in "
                f"map bounds '{self.nwse()}'.",
            ))

        # Determine the indices of the region in the latitude and longitude lists
        lat_indices = [
            i for i, lat in enumerate(self.latitude)
            if south <= lat <= north
        ]
        lon_indices = [
            i for i, lon in enumerate(self.longitude)
            if west <= lon <= east
        ]

        # Create a new map with the cropped coordinates
        return Success(dataclasses.replace(
            self,
            latitude=[self.latitude[i] for i in lat_indices],
            longitude=[self.longitude[i] for i in lon_indices],
        ))

    def nwse(self) -> tuple[float, float, float, float]:
        """Return the north, west, south, and east bounds of the map.

        Returns:
            A tuple of the north, west, south, and east bounds of the map.
        """
        if self.latitude is not None and self.longitude is not None:
            return self.latitude[0], self.longitude[0], self.latitude[-1], self.longitude[-1]
        return 90.0, -180.0, -90.0, 180.0 # TODO: Cross this bridge when we come to it




"""Implementation of adaptors for driven actors.

Driven actors
--------------

A driven actor is an external component that is acted upon by the core logic.
Also referred to as *secondary* actors, a driven actor represents an external
system that the core logic interacts with. They extend the core driven ports
(see `nwp_consumer.internal.ports`) in their implementation.

Examples of driven or secondary actors include:

- a database
- a message queue
- a filesystem

Since they are stores of data, they are referred to in this package
(and often in hexagonal architecture documentation) as *repositories*.

This module
-----------

This module contains implementations for the following driven actors:

- Notification Repository - Somewhere to send notifications to
- Model Repository - A source of NWP data

Both inherit from the repository ports specified in the core via `nwp_consumer.internal.ports`.
"""
from . import (
    raw_repositories,
    notification_repositories,
)

__all__ = [
    "raw_repositories",
    "notification_repositories",
]



import dataclasses
import datetime as dt
import os
import pathlib
import unittest
from unittest.mock import patch

from returns.result import Failure, ResultE, Success

from ...entities import NWPDimensionCoordinateMap, Parameter
from .mo_datahub import MetOfficeDatahubRawRepository


class TestMetOfficeDatahubRawRepository(unittest.TestCase):
    """Test the business methods of the MetOfficeDatahubRawRepository class."""

    @unittest.skipIf(
        condition="CI" in os.environ,
        reason="Skipping integration test that requires MetOffice DataHub access.",
    )
    def test__download(self) -> None:
        """Test the _download method."""

        auth_result = MetOfficeDatahubRawRepository.authenticate()
        self.assertIsInstance(auth_result, Success, msg=f"{auth_result!s}")
        c = auth_result.unwrap()

        test_it = c.repository().determine_latest_it_from(
            dt.datetime.now(tz=dt.UTC),
            c.model().running_hours,
        )

        dl_result = c._download(
            f"{c.request_url}/agl_u-component-of-wind-surface-adjusted_10.0_{test_it:%Y%m%d%H}_1/data",
        )

        self.assertIsInstance(dl_result, Success, msg=f"{dl_result!s}")

    def test_convert(self) -> None:
        """Test the _convert method."""

        @dataclasses.dataclass
        class TestCase:
            filename: str
            expected_coords: NWPDimensionCoordinateMap
            should_error: bool

        tests: list[TestCase] = [
            TestCase(
                filename="test_MODatahub_UM-Global_t2m_20241120T00_S00.grib",
                expected_coords=dataclasses.replace(
                    MetOfficeDatahubRawRepository.model().expected_coordinates,
                    init_time=[dt.datetime(2024, 11, 20, 0, tzinfo=dt.UTC)],
                    variable=[Parameter.TEMPERATURE_SL],
                    step=[0],
                ),
                should_error=False,
            ),
            TestCase(
                filename="test_MODatahub_UM-Global_u10_20241120T00_S17.grib",
                expected_coords=dataclasses.replace(
                    MetOfficeDatahubRawRepository.model().expected_coordinates,
                    init_time=[dt.datetime(2024, 11, 20, 0, tzinfo=dt.UTC)],
                    variable=[Parameter.WIND_U_COMPONENT_10m],
                    step=[17],
                ),
                should_error=False,
            ),
            TestCase(
                filename="test_HRES-IFS_10u.grib",
                expected_coords=MetOfficeDatahubRawRepository.model().expected_coordinates,
                should_error=True,
            ),
        ]

        for t in tests:
            with self.subTest(name=t.filename):
                # Attempt to convert the file
                result = MetOfficeDatahubRawRepository._convert_global(
                    path=pathlib.Path(__file__).parent.absolute() / "test_gribs" / t.filename,
                )
                region_result: ResultE[dict[str, slice]] = result.do(
                    region
                    for das in result
                    for da in das
                    for region in NWPDimensionCoordinateMap.from_xarray(da).bind(
                        t.expected_coords.determine_region,
                    )
                )
                if t.should_error:
                    self.assertIsInstance(region_result, Failure, msg=f"{region_result}")
                else:
                    self.assertIsInstance(region_result, Success, msg=f"{region_result}")

    @patch.dict(os.environ, {"MODEL": "um-ukv-2km"}, clear=True)
    def test_convert_ukv(self) -> None:


        @dataclasses.dataclass
        class TestCase:
            filename: str
            expected_coords: NWPDimensionCoordinateMap
            should_error: bool

        tests: list[TestCase] = [
            TestCase(
                filename="test_MO_UKV_agl_relative-humidity_1.5_2025012112.grib",
                expected_coords=dataclasses.replace(
                    MetOfficeDatahubRawRepository.model().expected_coordinates,
                    init_time=[dt.datetime(2025, 1, 21, 12, tzinfo=dt.UTC)],
                    variable=[Parameter.TEMPERATURE_SL],
                    step=[0],
                ),
                should_error=False,
            ),
        ]

        for t in tests:
            with self.subTest(name=t.filename):
                # Attempt to convert the file
                result = MetOfficeDatahubRawRepository._convert_ukv(
                    path=pathlib.Path(__file__).parent.absolute() / "test_gribs" / t.filename,
                )
                region_result: ResultE[NWPDimensionCoordinateMap] = result.do(
                    region
                    for das in result
                    for da in das
                    for region in NWPDimensionCoordinateMap.from_xarray(da)
                )
                if t.should_error:
                    self.assertIsInstance(region_result, Failure, msg=f"{region_result}")
                else:
                    self.assertIsInstance(region_result, Success, msg=f"{region_result}")



import unittest

from .postprocess import PostProcessOptions


class TestPostProcessOptions(unittest.TestCase):
    """Test the business methods of the PostProcessOptions class."""

    def test_requires_postprocessing(self) -> None:
        """Test that an empty initialization means no postprocessing."""
        test_class = PostProcessOptions()

        self.assertFalse(
            test_class.requires_postprocessing(),
            msg="Empty class should not require postprocessing.",
        )

    def test_requires_rewrite(self) -> None:
        """Test that an empty initialization means no rewriting."""
        test_class = PostProcessOptions()

        self.assertFalse(
            test_class.requires_rewrite(),
            msg="Empty class should not require rewriting.",
        )


if __name__ == "__main__":
    unittest.main()



import contextlib
import dataclasses
import datetime as dt
import logging
import os
import unittest
from collections.abc import Generator
from types import TracebackType
from unittest.mock import patch

import numpy as np
import xarray as xr
from botocore.client import BaseClient as BotocoreClient
from botocore.session import Session
from moto.server import ThreadedMotoServer
from returns.pipeline import is_successful
from returns.result import Success

from .coordinates import NWPDimensionCoordinateMap
from .parameters import Parameter
from .postprocess import PostProcessOptions
from .tensorstore import TensorStore

logging.getLogger("werkzeug").setLevel(logging.ERROR)


class MockS3Bucket:

    client: BotocoreClient
    server: ThreadedMotoServer
    bucket: str = "test-bucket"

    def __enter__(self) -> None:
        """Create a mock S3 server and bucket."""
        self.server = ThreadedMotoServer()
        self.server.start()

        session = Session()
        self.client = session.create_client(
            service_name="s3",
            region_name="us-east-1",
            endpoint_url="http://localhost:5000",
            aws_access_key_id=os.environ["AWS_ACCESS_KEY_ID"],
            aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"],
        )

        self.client.create_bucket(
            Bucket=self.bucket,
        )

    def __exit__(
            self,
            exc_type: type[BaseException] | None,
            exc_val: BaseException | None,
            exc_tb: TracebackType | None,
        ) -> None:
        response = self.client.list_objects_v2(
            Bucket=self.bucket,
        )
        """Delete all objects in the bucket and stop the server."""
        if "Contents" in response:
            for obj in response["Contents"]:
                self.client.delete_object(
                    Bucket=self.bucket,
                    Key=obj["Key"],
                )
        self.server.stop()


class TestTensorStore(unittest.TestCase):
    """Test the business methods of the TensorStore class."""

    @contextlib.contextmanager
    def store(self, year: int) -> Generator[TensorStore, None, None]:
        """Create an instance of the TensorStore class."""

        test_coords: NWPDimensionCoordinateMap = NWPDimensionCoordinateMap(
            init_time=[
                dt.datetime(year, 1, 1, h, tzinfo=dt.UTC)
                for h in [0, 6, 12, 18]
            ],
            step=[1, 2, 3, 4],
            variable=[Parameter.TEMPERATURE_SL],
            latitude=np.linspace(90, -90, 12).tolist(),
            longitude=np.linspace(0, 360, 18).tolist(),
        )

        init_result = TensorStore.initialize_empty_store(
            model="test_da",
            repository="dummy_repository",
            coords=test_coords,
            chunks=test_coords.chunking(chunk_count_overrides={}),
        )
        self.assertIsInstance(init_result, Success, msg=init_result)
        store = init_result.unwrap()
        yield store
        store.delete_store()

    @patch.dict(os.environ, {
        "AWS_ENDPOINT_URL": "http://localhost:5000",
        "AWS_ACCESS_KEY_ID": "test-key",
        "AWS_SECRET_ACCESS_KEY": "test-secret",
        "ZARRDIR": "s3://test-bucket/data",
    }, clear=True)
    def test_initialize_and_delete_s3(self) -> None:
        """Test the initialize_empty_store method."""

        with MockS3Bucket(), self.store(year=2022) as ts:
            delete_result = ts.delete_store()
            self.assertIsInstance(delete_result, Success, msg=delete_result)

    def test_write_to_region(self) -> None:
        """Test the write_to_region method."""
        with self.store(year=2022) as ts:
            test_da: xr.DataArray = xr.DataArray(
                name="test_da",
                data=np.ones(
                    shape=list(ts.coordinate_map.shapemap.values()),
                ),
                coords=ts.coordinate_map.to_pandas(),
            )

            # Write each init time and step one at a time
            for it in test_da.coords["init_time"].values:
                for step in test_da.coords["step"].values:
                    write_result = ts.write_to_region(
                        da=test_da.where(
                            test_da["init_time"] == it, drop=True,
                        ).where(test_da["step"] == step, drop=True),
                    )
                    self.assertIsInstance(write_result, Success, msg=write_result)

    def test_postprocess(self) -> None:
        """Test the postprocess method."""

        @dataclasses.dataclass
        class TestCase:
            name: str
            options: PostProcessOptions
            should_error: bool

        tests: list[TestCase] = [
            TestCase(
                name="empty_options",
                options=PostProcessOptions(),
                should_error=False,
            ),
        ]

        with self.store(year=1971) as ts:
            for t in tests:
                with self.subTest(name=t.name):
                    result = ts.postprocess(t.options)
                    if t.should_error:
                        self.assertTrue(
                            isinstance(result, Exception),
                            msg="Expected error to be returned.",
                        )
                    else:
                        self.assertTrue(is_successful(result))

    def test_missing_times(self) -> None:
        """Test the missing_times method."""

        @dataclasses.dataclass
        class TestCase:
            name: str
            times_to_write: list[dt.datetime]
            expected: list[dt.datetime]

        with self.store(year=2024) as ts:
            tests: list[TestCase] = [
                TestCase(
                    name="all_missing_times",
                    times_to_write=[],
                    expected=ts.coordinate_map.init_time,
                ),
                TestCase(
                    name="some_missing_times",
                    times_to_write=[ts.coordinate_map.init_time[0], ts.coordinate_map.init_time[2]],
                    expected=[ts.coordinate_map.init_time[1], ts.coordinate_map.init_time[3]],
                ),
                TestCase(
                    name="no_missing_times",
                    times_to_write=ts.coordinate_map.init_time,
                    expected=[],
                ),
            ]

            for t in tests:
                with self.subTest(name=t.name):
                    for i in t.times_to_write:
                        write_result = ts.write_to_region(
                            da=xr.DataArray(
                                name="test_da",
                                data=np.ones(
                                    shape=[
                                        1 if k == "init_time" else v
                                        for k, v in ts.coordinate_map.shapemap.items()
                                    ],
                                ),
                                coords=ts.coordinate_map.to_pandas() | {
                                    "init_time": [np.datetime64(i.replace(tzinfo=None), "ns")],
                                },
                            ),
                        )
                        write_result.unwrap()
                    result = ts.missing_times()
                    missing_times = result.unwrap()
                    self.assertListEqual(missing_times, t.expected)

if __name__ == "__main__":
    unittest.main()

